use crate::core::input::InputManager;
use crate::core::render::pipelines::triangle_pipeline::create_triangle_pipeline;
use crate::core::render::primitives::camera::Camera;
use crate::core::render::primitives::vertex::Vertex2D;
use crate::core::render::render_context::RenderContext;
use crate::core::scene::Scene;
use crate::core::timer::Timer;
use glam::{Mat4, Quat, Vec3};
use std::sync::Arc;
use vulkano::buffer::Subbuffer;
use vulkano::command_buffer::{AutoCommandBufferBuilder, PrimaryAutoCommandBuffer};
use vulkano::descriptor_set::{DescriptorSet, WriteDescriptorSet};
use vulkano::pipeline::{GraphicsPipeline, Pipeline, PipelineBindPoint};

const VERTICES: [Vertex2D; 12] = [
    // Triangle en haut à gauche
    Vertex2D {
        position: [-0.5, -0.75],
        color: [1.0, 0.0, 0.0],
    },
    Vertex2D {
        position: [-0.75, -0.25],
        color: [0.0, 1.0, 0.0],
    },
    Vertex2D {
        position: [-0.25, -0.25],
        color: [0.0, 0.0, 1.0],
    },
    // Triangle en bas à gauche
    Vertex2D {
        position: [-0.5, 0.25],
        color: [0.5, 0.5, 0.5],
    },
    Vertex2D {
        position: [-0.75, 0.75],
        color: [0.2, 0.8, 0.2],
    },
    Vertex2D {
        position: [-0.25, 0.75],
        color: [0.8, 0.2, 0.2],
    },
    // Triangle en haut à droite
    Vertex2D {
        position: [0.5, -0.75],
        color: [1.0, 1.0, 0.0],
    },
    Vertex2D {
        position: [0.25, -0.25],
        color: [0.0, 1.0, 1.0],
    },
    Vertex2D {
        position: [0.75, -0.25],
        color: [1.0, 0.0, 1.0],
    },
    // Triangle en bas à droite
    Vertex2D {
        position: [0.5, 0.25],
        color: [0.1, 0.5, 0.8],
    },
    Vertex2D {
        position: [0.25, 0.75],
        color: [0.8, 0.6, 0.1],
    },
    Vertex2D {
        position: [0.75, 0.75],
        color: [0.3, 0.4, 0.6],
    },
];

pub struct MainSceneState {
    pipeline: Arc<GraphicsPipeline>,
    vertex_buffer: Subbuffer<[Vertex2D]>,
    camera: Camera,
}

#[derive(Default)]
pub struct MainScene {
    state: Option<MainSceneState>,
}

impl Scene for MainScene {
    fn loaded(&self) -> bool {
        self.state.is_some()
    }

    fn load(&mut self, render_context: &RenderContext) {
        let pipeline =
            create_triangle_pipeline(render_context.device(), render_context.swapchain_format())
                .unwrap();
        let vertex_buffer =
            Vertex2D::create_buffer(Vec::from_iter(VERTICES), render_context.memory_allocator())
                .unwrap();

        let camera = Camera::new(
            Mat4::look_at_rh(
                Vec3::new(0.3, 0.3, 1.0),
                Vec3::new(0.0, 0.0, 0.0),
                Vec3::new(0.0, -1.0, 0.0),
            ),
            Mat4::perspective_rh_gl(
                std::f32::consts::FRAC_PI_2,
                render_context.aspect_ratio(),
                0.01,
                100.0,
            ),
        );

        self.state = Some(MainSceneState {
            pipeline,
            vertex_buffer,
            camera,
        })
    }

    fn update(
        &mut self,
        render_context: &RenderContext,
        input_manager: &InputManager,
        timer: &Timer,
    ) {
        let state = self.state.as_mut().unwrap();

        let speed = 50.0 * timer.delta_time();

        let mut rot = Quat::default();
        rot *= Quat::from_rotation_y(
            input_manager.get_virtual_input_state("mouse_x") * speed.to_radians(),
        );
        rot *= Quat::from_rotation_x(
            input_manager.get_virtual_input_state("mouse_y") * speed.to_radians(),
        );
        state.camera.get_transform_mut().rotate(rot);

        let translation_x =
            input_manager.get_virtual_input_state("move_right") * timer.delta_time() * speed;

        let translation_z =
            input_manager.get_virtual_input_state("move_forward") * timer.delta_time() * speed;

        state
            .camera
            .get_transform_mut()
            .translate(Vec3::new(translation_x, 0.0, translation_z));

        state.camera.set_projection(Mat4::perspective_rh_gl(
            std::f32::consts::FRAC_PI_2,
            render_context.aspect_ratio(),
            0.01,
            100.0,
        ));
    }

    fn render(
        &self,
        render_context: &RenderContext,
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) {
        let state = self.state.as_ref().unwrap();
        let vertex_count = state.vertex_buffer.len() as u32;
        let instance_count = vertex_count / 3;

        let layout = &state.pipeline.layout().set_layouts()[0];
        let uniform_buffer = state
            .camera
            .create_buffer(render_context.memory_allocator())
            .unwrap();
        let descriptor_set = DescriptorSet::new(
            render_context.descriptor_set_allocator().clone(),
            layout.clone(),
            [WriteDescriptorSet::buffer(0, uniform_buffer)],
            [],
        )
        .unwrap();

        unsafe {
            builder
                .bind_pipeline_graphics(state.pipeline.clone())
                .unwrap()
                .bind_descriptor_sets(
                    PipelineBindPoint::Graphics,
                    state.pipeline.layout().clone(),
                    0,
                    descriptor_set,
                )
                .unwrap()
                .bind_vertex_buffers(0, state.vertex_buffer.clone())
                .unwrap()
                .draw(vertex_count, instance_count, 0, 0)
                .unwrap();
        }
    }

    fn unload(&mut self) {}
}
